{$mode delphi}
unit TomasProcess;

(******************************************************************)
(* revision history:                                              *)
(*  26-8-2001:  StandBy solved, eventcounter added                *)
(*              Future event added                                *)
(*  03-4-2001:  Question/answer mechanism added                   *)
(*  02-4-2001:  Standby mechanisme verbeterd                      *)
(*              messagehandling met server uitgebreid en aangepast*)
(*  26-9-2000:  Tomasserver methods implemented                   *)
(*  15-8-2000:  ResetTomasProcess added.                          *)
(*  20-7-2000:  concurrent development facilities implemented     *)
(*              and tested.                                       *)
(*  14-6-2000:  entersequence corrected                           *)
(*              reordered standby's kept original eventtime       *)
(*   6-6-2000:  standby corrected.                                *)
(*              Finishanddestroy corrected                        *)
(******************************************************************)


interface

  Uses
    Classes,
    Windows,
    Forms,
    SysUtils,
    Dialogs,
    Controls;

Const
  NEVER: Double = 1.7E+307;
  ProcessVersion: String = 'V2.00.1';

Type

  ChangeProcedure = Procedure;
  SimulationStatus = (S_INIT,S_FINISHED,S_INTERRUPTED,S_RUNNING,S_STEPRUNNING);
  ElementStatus = (TC_DATA,TC_CLAIMING,TC_PASSIVE,TC_INTERRUPTED,
   TC_TIMESCHEDULED,TC_STATESCHEDULED,TC_CURRENT);

  {TomasProcessElement is the basic class of TomasProcess.
   Active Elements should override the process-method.
   Remember to put this method in the 'published' section
   of your Element-definition.}
  TomasProcessElement = class(TPersistent)
    Private
      FName: String;
      FReturnAddress: Pointer;
      FLocData: Pointer;
      FStatus: ElementStatus;
      FArrivalTime: Double;
      FEventTime: Double;
    Public
      Claims: TList;
      Property Name: String read FName write FName;
      Property Status: ElementStatus read FStatus write FStatus;
      Property Arrivaltime: Double read FArrivalTime write FArrivalTime;
      Property EventTime: Double read FEventTime write FEventTime;
      Destructor Destroy; override;
    Published
      Constructor Create(S: String; CTime: Double);

      {The process method contains a complete
       process description of the Element.}
      {example:
       In this example the process of a machine
       is described. A machine is a TomasProcessElement
       with some task(s) to be performed. Tasks are also
       TomasProcessElements that have an attribute TaskTime.
       Tasks to be done are put in the machine's
       taskqueue by some other Element. When there
       are no tasks left, the machine suspends.

       Procedure MachineElement.Process;
         Var
           Task: TaskElement;
         Begin
         While TRUE Do
           Begin
           While TaskQueue.GetLength > 0 Do
             Begin
             Task:=TaskQueue.GetFirstElement;
             TaskQueue.Remove(Task);
             HoldProcess(Task.TaskTime);
             End;
           SuspendProcess;
           End;
         End;
      }
      Procedure Process; virtual;
      Procedure Receive; virtual;
      Procedure ScheduleProcess(T: Double);
      Procedure RemoveProcess(Stop: Boolean);
      Procedure InterruptProcess;
      Procedure CancelProcess;
      procedure CancelResume;
      {Scheduled returns TRUE if the element is
      scheduled by the sequencing mechanism.}
      Function Scheduled: Boolean;
    End;




Procedure HoldProcess(T: Double);
Procedure StandByProcess;
Procedure SuspendProcess;
Procedure FinishProcess;
Procedure FinishProcessAndDestroy;


Procedure StartProcessSimulation;
Procedure ResumeProcessSimulation;
Procedure InterruptProcessSimulation;
Procedure GoSetSimulationSpeed(Speed: Integer);
Procedure SetStatusChange(ChangeProc: ChangeProcedure);
Function CurrentProcessElement: TomasProcessElement;
Procedure SetKeyPressed(C: Char);
Function TracingMessage: Boolean;
Function TraceLine: String;
Function CurrentChange: Boolean;
//Function ServerMessage: Boolean;
//Function ServerQuestion: Boolean;
//Procedure QuestionParms(Var Question,Source: String; Var ID: Integer);
//procedure SendModelMessage(Model,Msg: String);
//Function MessageLine: String;
Procedure GoSetModelName(Name: String);
//Procedure SetServer(Srv: String);
//Function GetServer: String;
Procedure SetBreakTime(T: Double);
Function GetBreakTime: Double;
Procedure SetStopTime(T: Double);
Function GetStopTime: Double;
Function GetSimulationStatus: SimulationStatus;
Procedure StartStepMode;
Procedure StopStepMode;
Procedure PerformNextEvent;
Function GetNow: Double;
Procedure ResetTomasProcess;
//Procedure CheckServer;
//Procedure DisconnectFromServer;
Procedure DoStartLogging;
Procedure DoStopLogging;
Procedure SetProcessingInterval(PInt: Integer);
Function GetLogging: Boolean;
//Procedure DoSendQuestionTo(Model: String; Var Question: String);
//Procedure DoSendReplyTo(Model,Reply: String; ID: Integer);
//Function ConnectedToServer: Boolean;
Procedure DoSetSynchronization(OnOff: Boolean);

implementation

Type

  TQuestion = class
    private
    FID: Integer;
    FQuestion: String;
    FTarget: String;
    public
    property Question: String read FQuestion;
    property Destination: String read FTarget;
    property Source: String read FTarget;
    End;

  TReply = class
    private
    FID: Integer;
    FSource: String;
    FAnswer: String;
    public
    property Answer: String read FAnswer;
    property Source: String read FSource;
    property Destination: String read FSource;
    End;

  TTomasStatus = class
    TS_SimulationStatus: SimulationStatus;
    TS_NextStep: Boolean;
    TS_CurrentElement: TomasProcessElement;
    TS_ServerAddress: String;
    TS_Now: Double;
    TS_Tracing: String;
    TS_Message: String;
    TS_CurrentChange: Boolean;
    TS_Question: TQuestion;
    TS_SimulationSpeed: Integer;
    Constructor Create;
    End;

  LocalData = ^LocalDataRecord; {structure created by: hold, }
  LocalDataRecord =             {Suspend or standby        }
    Record
    LD_Length: LongInt; {must have been hold/startat etc. followed }
    LD_Data: Array[1..64000] Of Byte; {by a cancel}
    End;


  Sequel = ^SequelRecord;
  SequelRecord =
    Record
    SR_Next: Sequel;
    SR_Pred: Sequel;
    SR_Element: TomasProcessElement;
    SR_Time: Double;
    SR_StandBy: Boolean;
    SR_LocalData: Localdata;
    End;
(*
  TModelSocket = class(TWSocket)
    published
    Constructor Create(AOwner: TComponent); override;
    procedure RegisterConnection(Sender: TObject; Error: Word);
    procedure CheckServerAnswer(Sender: TObject; Error: Word);
    End;
*)
//Const
//  ServerMessages : Array[0..10] of string = ('','$$!','$$?','$$hello:','$$continue:',
//   '$$stop:','$$simtime:','$$time:','$$clients:','$$from:','$$unknown:');

Var
  TomasStatus: TTomasStatus;
  SequentialMechanism: Sequel;
  StandByMechanism: Sequel;
  FutureEvent: Sequel;
  InStandBy: Boolean;
  SB_Successor: Sequel;
  CurrentSequel: Sequel;
  CurElement: TomasProcessElement;
  TheElement: TomasProcessElement;
  DestroyRoom: TList;
  QuestionList: TList;
  QuestionID: Integer;
  ReplyList: TList;
  NoReplyReceived: Boolean;
  ReplyRecognized: Boolean;
  StackPosition: Pointer;
  JumpAddress: Pointer;
  NrOfBytes: LongInt;
  InitialStack: Pointer;
  ReturnAddress: Pointer;
  LocBytes: Integer;
  LocData: LocalData;
  StartEAX: Integer;
  StartEBX: Integer;
  StartECX: Integer;
  StartESI: Integer;
  StartEDI: Integer;
  StartEBP: Integer;
  SimulationStopTime: Double;
  GoOnModel: Boolean;
  SimulationBreakTime: Double;
  SimulationKeyTime: Double;
  SimulationReturn: Pointer;
  StatusChange: ChangeProcedure;
  InStepMode: Boolean;
//  ModelSocket: TWSocket;
//  ModelNumber: String;
  ModelName: String;
//  WaitMsg: String;
//  MessageReceived: Boolean;
  Buffer: Array[0..2047] Of Char;
  Logging: Boolean;
  LoggingStarted: Boolean;
  LogFile: TextFile;
  LogCount: Integer;
  EventCounter: Integer;
  ProcessingInterval: Integer;
  Synchronization: Boolean;
  
Procedure SelectNextCurrent; forward;
Function CompInSequence(Comp: TomasProcessElement): Boolean; forward;


procedure LogMessage(Msg,MsgType: String);
  Var
    Log2File: TextFile;
    S: String;
  begin
  If Logging Then
    Begin
    Inc(LogCount);
    If LogCount >= 1000 Then
      Begin
      Reset(LogFile);
      While LogCount > 500 Do
        Begin
        ReadLn(LogFile,S);
        Dec(LogCount);
        End;
      AssignFile(Log2File,ModelName + '.lg2');
      ReWrite(Log2File);
      While Not(EOF(LogFile)) Do
        Begin
        ReadLn(LogFile,S);
        WriteLn(Log2File,S);
        End;
      CloseFile(LogFile);
      CloseFile(Log2File);
      Erase(LogFile);
      Rename(Log2File,ModelName + '.log');
      End;
    Append(LogFile);
    WriteLn(LogFile,FormatFloat('0.00',GetNow) + MsgType + Msg);
    CloseFile(LogFile);
    End;
  End;

Procedure DoSetSynchronization(OnOff: Boolean);
  Begin
  Synchronization:=OnOff;
  End;  

Constructor TTomasStatus.Create;
  Begin
  inherited;
  TS_SimulationStatus:=S_INIT;
  TS_CurrentElement:=Nil;
  TS_ServerAddress:='';
  TS_Now:=0;
  TS_Tracing:='';
  TS_Message:='';
  TS_CurrentChange:=FALSE;
  TS_Question:=Nil;
  TS_SimulationSpeed:=100;
  End;  {}

Procedure GoSetSimulationSpeed(Speed: Integer);
  Begin
  If Speed < 0 Then Speed:=0;
  If Speed > 100 Then Speed:=100;
  TomasStatus.TS_SimulationSpeed:=Speed;
  End;


Procedure SetKeyPressed(C: Char);
  Begin
  If C = ' ' Then
    Begin
    SimulationKeyTime:=TomasStatus.TS_Now;
    End;
  End;

Function TracingMessage: Boolean;
  Begin
  Result:=(TomasStatus.TS_Tracing<>'');
  End; {TracingMessage}

Function TraceLine: String;
  begin
  TraceLine:=TomasStatus.TS_Tracing;
  TomasStatus.TS_Tracing:='';
  End;  {TraceLine}

Function CurrentChange: Boolean;
  Begin
  Result:=TomasStatus.TS_CurrentChange;
  End;


(*
Function ServerMessage: Boolean;
  Begin
  Result:=(TomasStatus.TS_Message<>'');
  End;

Function ServerQuestion: Boolean;
  Begin
  Result:=(TomasStatus.TS_Question<>Nil);
  End;

Procedure QuestionParms(Var Question,Source: String; Var ID: Integer);
  Begin
  Question:=TomasStatus.TS_Question.Question;
  Source:=TomasStatus.TS_Question.Source;
  ID:=TomasStatus.TS_Question.FID;
  TomasStatus.TS_Question.Free;
  TomasStatus.TS_Question:=Nil;
  End;


Procedure TomasSendStr(S: String);
  Var
    k: Integer;
    i: Integer;
  Begin
  LogMessage(S,' sent: ');
  i:=0;
  k:=ModelSocket.SendStr(S);
  While (k = -1) And (i<10) Do
   Begin
   k:=ModelSocket.SendStr(S);
   inc(i);
   End;
  If (k = -1) And (i=10) Then
    begin
    If Logging Then
      Begin
      LogMessage('SendStr from model '+ ModelName,' Error: ');
      End
    Else
      Begin
      MessageDlg('Error sending message: '+ S,mtWarning,[mbOK],0);
      End;
    End;
  End;

procedure SendModelMessage(Model,Msg: String);
  Begin
  If Model = '$$server' Then
    Begin
    TomasSendStr('$$' + Msg + #13#10);
    End
  Else
    Begin
    TomasSendStr('$$toclient: ' + Model + '!!' + Msg + #13#10);
    End;
  End;

Function MessageLine: String;
  Begin
  MessageLine:=TomasStatus.TS_Message;
  TomasStatus.TS_Message:='';
  End;
*)



Procedure SetStatusChange(ChangeProc: ChangeProcedure);
  Begin
  StatusChange:=ChangeProc;
  End;

Procedure GoSetModelName(Name: String);
  Begin
  ModelName:=Name;
  End;



(*
Procedure SetServer(Srv: String);
  Begin
  TomasStatus.TS_ServerAddress:=Trim(Srv);
  End;

Function GetServer: String;
  Begin
  Result:=TomasStatus.TS_ServerAddress;
  ENd;
*)



Procedure SetBreakTime(T: Double);
  Begin
  SimulationBreakTime:=T;
  End;

Function GetBreakTime: Double;
  Begin
  Result:=SimulationBreakTime;
  End;

Procedure SetStopTime(T: Double);
  Begin
  SimulationStopTime:=T;
  End;

Function GetStopTime: Double;
  Begin
  Result:=SimulationStopTime;
  End;

Function GetSimulationStatus: SimulationStatus;
  Begin
  Result:=TomasStatus.TS_SimulationStatus;
  End;

Procedure StartStepMode;
  Begin
  TomasStatus.TS_SimulationStatus:=S_STEPRUNNING;
  TomasStatus.TS_NextStep:=FALSE;
  InStepMode:=TRUE;
  End;

Procedure StopStepMode;
  Begin
  InStepMode:=FALSE;
  If TomasStatus.TS_SimulationStatus > S_FINISHED Then
    Begin
    TomasStatus.TS_SimulationStatus:=S_RUNNING;
    TomasStatus.TS_NextStep:=TRUE;
    End;
  End;

Procedure PerformNextEvent;
  Begin
  TomasStatus.TS_NextStep:=TRUE;
  End;

Function GetNow: Double;
  Begin
  Result:=TomasStatus.TS_Now;
  End;


Function CurrentTimeString: String;
  Var
    S: String;
  Begin
  Str(TomasStatus.TS_Now:10:2,S);
  Result:=S;
  End;   {CurrentTimeString}


Constructor TomasProcessElement.Create(S: String; CTime: Double);
  Begin
  inherited Create;
  FName:=S;
  FStatus:=TC_DATA;
  FArrivalTime:=CTime;
  FEventTime:=Never;
  FLocData:=Nil;
  Claims:=TList.Create;
  End;

Destructor TomasProcessElement.Destroy;
  Begin
  Claims.Clear;
  Claims.Free;
  inherited Destroy;
  End;

Procedure TomasProcessElement.Process;
  Begin
  MessageDlg(CurrentTimeString + ' Warning: ' + Name + ' has no process defined',
   mtWarning,[mbOK],0);
  SelectNextCurrent;
  End;      {TomasProcessElement.Process}

Procedure TomasProcessElement.Receive;
  Begin
  MessageDlg(CurrentTimeString + ' Warning: ' + Name + ' has no receive method defined',
   mtWarning,[mbOK],0);
  End;

Function TomasProcessElement.Scheduled: Boolean;
  Begin
  Result:=CompInSequence(Self);
  End;

Procedure DoContinue;
  Begin
//  CurrentSequel:=SequentialMechanism^.SR_Next;
  CurrentSequel^.SR_Element.FStatus:=TC_Current;
  {Do NOT set an attribute of _CurrentElement.
   The stack will be influenced by that!!!!!}
  CurElement:=CurrentSequel^.SR_Element;
  TomasStatus.TS_CurrentElement:=CurElement;
  If NOT(InStandBy) Then
    Begin
    If TomasStatus.TS_Now <> CurrentSequel^.SR_Time Then
      Begin
      TomasStatus.TS_Now:=CurrentSequel^.SR_Time;
      FutureEvent:=CurrentSequel^.SR_Next;
      End;
    End;
  If FutureEvent = CurrentSequel Then
    FutureEvent:=CurrentSequel^.SR_Next;   
  CurElement.FEventTime:=Never;
  ReturnAddress:=CurElement.FReturnAddress;
  TomasStatus.TS_CurrentChange:=TRUE;
  //TomasStatus.TS_Tracing:=TimeToStr(Time) + '/' + CurrentTimeString + '  ' +
  // CurElement.Name + ' is current now';
  StatusChange;
  TomasStatus.TS_CurrentChange:=FALSE;
  CurrentSequel^.SR_Pred^.SR_Next:=CurrentSequel^.SR_Next;
  CurrentSequel^.SR_Next^.SR_Pred:=CurrentSequel^.SR_Pred;
  If CurrentSequel^.SR_LocalData<>Nil Then
    Begin
    LocBytes:=CurrentSequel^.SR_LocalData^.LD_length;
    LocData:=Addr(CurrentSequel^.SR_LocalData^.LD_Data[1]);
      asm
      mov  esi,LocData
      mov  ecx,LocBytes
      mov  edi,InitialStack
      sub  edi,LocBytes
      mov  esp,edi
      rep  movsb      {restore stack-values}
     {registers have been put there by hold/suspend/standby}
      end;
    FreeMem(CurrentSequel^.SR_LocalData,CurrentSequel^.SR_LocalData^.LD_Length+4);
    CurrentSequel^.SR_Localdata:=Nil;
    LocData:=Nil;
    End
  Else             {start of process}
    Begin
      asm
      mov esp,InitialStack
      mov ebp,StartEBP
      push edi
      push esi
      push ebx
      push ebp
      end;
    End;
 Dispose(CurrentSequel);
 {Now  jump to returnaddress i.e RetHold if started from hold}
   asm
   pop  ebp
   pop  ebx
   pop  esi
   pop  edi
   mov eax,CurElement   {eax must point to comp}
   mov edx,ReturnAddress
   jmp edx
   end;
  End;  {DoContinue}

Procedure AskToContinue;
  Begin
//  TomasSendStr('$$continue:' + FormatFloat('0.00',CurrentSequel^.SR_Time) + #13#10);
  GoOnModel:=FALSE;
  While NOT(GoOnModel) Do
    Begin
    Application.ProcessMessages;
    End;
  DoContinue;
  End;

procedure NotifyServer;
  Begin
  If (TomasStatus.TS_ServerAddress <> '')
   And (TomasStatus.TS_CurrentElement = Nil)
   And (TomasStatus.TS_SimulationStatus > S_Finished) Then
    Begin
    CurrentSequel:=SequentialMechanism^.SR_Next;
    If (CurrentSequel^.SR_Time >= SimulationStopTime) Then
      Begin
//      TomasSendStr('$$newevent:' + FormatFloat('0.00',SimulationStopTime) + #13#10);
      End
    Else
      Begin
//       TomasSendStr('$$newevent:' + FormatFloat('0.00',CurrentSequel^.SR_Time) + #13#10);
      End;
    End;
  End;

(*
Procedure WaitForFinish;
  Begin
  If ModelSocket.State <> wsClosed Then
    Begin
    TomasSendStr('$$continue:' + FormatFloat('0.00',SimulationStopTime) + #13#10);
    GoOnModel:=FALSE;
    While NOT(GoOnModel) Do
      Begin
      Application.ProcessMessages;
      End;
    SelectNextCurrent;
    End;
  End;
*)

Procedure SelectNextCurrent;
  Begin
  TomasStatus.TS_CurrentElement:=Nil;
  While DestroyRoom.Count > 0 Do
    Begin
    TheElement:=DestroyRoom.Items[0];
    DestroyRoom.Remove(TheElement);
    TheElement.Destroy;
    End;
  If StandByMechanism^.SR_Next = StandByMechanism Then
    Begin
    InStandBy:=FALSE;
    CurrentSequel:=SequentialMechanism^.SR_Next;
    SB_Successor:=Nil;
    End
  Else   // there are standbys
    Begin
    If NOT(InStandBy) Then
      Begin
      CurrentSequel:=StandByMechanism^.SR_Next;
      InStandBy:=TRUE;
      SB_Successor:=CurrentSequel^.SR_Next;
      End
    Else
      Begin
      If SB_Successor = StandByMechanism Then  // all standbys checked
        Begin
        InStandBy:=FALSE;
        SB_Successor:=Nil;
        CurrentSequel:=SequentialMechanism^.SR_Next;
        End
      Else
        Begin
        CurrentSequel:=SB_Successor;
        SB_Successor:=CurrentSequel^.SR_Next;
        End;
      End;
    If InStandBy Then
      begin
      CurrentSequel^.SR_Time:=TomasStatus.TS_Now;
      End;
    End;
  If ((SimulationBreakTime<>-1) And
   (CurrentSequel^.SR_Time >= SimulationBreakTime)) Or (SimulationKeyTime<>-1) Then
    Begin
    If SimulationKeyTime <> -1 Then
      Begin
      TomasStatus.TS_Now:=SimulationKeyTime;
      SimulationKeyTime:=-1;
      End
    Else
      Begin
      TomasStatus.TS_Now:=SimulationBreakTime;
      SimulationBreakTime:=-1;
      End;
    TomasStatus.TS_SimulationStatus:=S_INTERRUPTED;
    StatusChange;
    End;
  If TomasStatus.TS_SimulationStatus = S_STEPRUNNING Then
    Begin
    TomasStatus.TS_NextStep:=FALSE;
    End;
  Dec(EventCounter);
  If EventCounter = 0 Then
    Begin
    Application.ProcessMessages;
    EventCounter:=ProcessingInterval;
    End;
  While ((TomasStatus.TS_SimulationStatus < S_RUNNING) And
   (TomasStatus.TS_SimulationStatus <> S_FINISHED)) OR
    NOT(TomasStatus.TS_NextStep) Do
    Begin
    Application.ProcessMessages;
    End;
//  CurrentSequel:=SequentialMechanism^.SR_Next;
  If (CurrentSequel^.SR_Time >= SimulationStopTime) Then
    Begin

(*
    If (TomasStatus.TS_ServerAddress<>'') And Synchronization
     And (TomasStatus.TS_Now < SimulationStopTime) Then
      Begin
        asm
        pop  edx     {destroy ebx}
        pop  edx     {destroy returnaddress}
        end;
      WaitForFinish;
      End;
*)

    If TomasStatus.TS_SimulationStatus >= S_RUNNING Then
      Begin
      If SimulationStopTime >= 0 Then
        Begin
        TomasStatus.TS_Now:=SimulationStopTime;
        End;
      TomasStatus.TS_SimulationStatus:=S_FINISHED;
      StatusChange;
        asm
        pop edx   {pop ebx}
        pop edx   {pop returnaddress}
        mov esp,InitialStack
        mov ebp,StartEBP
        mov ebx,StartEBX
        mov esi,StartESI
        mov edi,StartEDI
        mov eax,STARTEAX
        mov ecx,STARTECX
        mov edx,SimulationReturn
        jmp edx
        end;
      End
    Else
      Begin
      Exit;
      End;
    End;
  If TomasStatus.TS_SimulationSpeed < 100 Then
    Sleep(100 - TomasStatus.TS_SimulationSpeed);
  If (TomasStatus.TS_ServerAddress <> '') And Synchronization Then
    Begin
    AskToContinue;
    End
  Else
    Begin
    DoContinue;
    End;
  End;   {SelectNextCurrent}

Procedure EnterSequence(T: Double; Comp: TomasProcessElement);
  Var
    NewSequence: Sequel;
    NextSequence: Sequel;
  Begin
  New(NewSequence);
  NewSequence^.SR_Element:=Comp;
  NewSequence^.SR_Time:=T;
  NewSequence^.SR_StandBy:=FALSE;
  NewSequence^.SR_LocalData:=LocData;
  LocData:=Nil;
  Comp.FStatus:=TC_TIMESCHEDULED;
  Comp.FEventTime:=T;
  If FutureEvent^.SR_Time > T Then
    Begin
    NextSequence:=FutureEvent;
    If T > GetNow Then
      FutureEvent:=NewSequence;
    End
  Else
    Begin
    NextSequence:=FutureEvent^.SR_Next;
    While (NextSequence^.SR_Time<=T) Do
      Begin
      NextSequence:=NextSequence^.SR_Next;
      End;
    End;
  NewSequence^.SR_Next:=NextSequence;
  NewSequence^.SR_Pred:=NextSequence^.SR_Pred;
  NewSequence^.SR_Pred^.SR_Next:=NewSequence;
  NextSequence^.SR_Pred:=NewSequence;
//  NotifyServer;
  End;   {EnterSequence}

Procedure EnterStandBySequence(Comp: TomasProcessElement; NewStandBy: Boolean);
//If NewStandBy then to end of StandByChain
// else remain in standbychain
  Var
    NewSequence: Sequel;
  Begin
  New(NewSequence);
  NewSequence^.SR_Element:=Comp;
  NewSequence^.SR_LocalData:=LocData;
  NewSequence^.SR_StandBy:=TRUE;
  NewSequence^.SR_Time:=GetNow;
  LocData:=Nil;
  Comp.FStatus:=TC_STATESCHEDULED;
  If NewStandBy Then
    Begin
    NewSequence^.SR_Next:=StandByMechanism;
    NewSequence^.SR_Pred:=StandByMechanism^.SR_Pred;
    NewSequence^.SR_Pred^.SR_Next:=NewSequence;
    StandByMechanism^.SR_Pred:=NewSequence;
    End
  Else
    Begin
    NewSequence^.SR_Next:=SB_Successor;
    NewSequence^.SR_Pred:=SB_Successor^.SR_Pred;
    SB_Successor^.SR_Pred:=NewSequence;
    NewSequence^.SR_Pred^.SR_Next:=NewSequence;
    End;
  Comp.FEventTime:=GetNow;
//  NotifyServer;
  End; {EnterStandBySequence}


Procedure TomasProcessElement.ScheduleProcess(T: Double);
  Begin
  If FLocData<>Nil Then
    Begin
    LocData:=FLocData;
    FLocData:=Nil;
    End
  Else
    Begin
    LocData:=Nil;
    FReturnAddress:=MethodAddress('Process');
    End;
  If T >= GetNow Then
    begin
    EnterSequence(T,Self);
    End
  Else
    Begin
    //after interrupt-period resume standby, so put element to end of standbychain
    EnterStandBySequence(Self,TRUE);
    End;
  End;


Procedure LeaveSequence(Comp: TomasProcessElement);
  Var
    NextSequence: Sequel;
  Begin
  NextSequence:=StandbyMechanism^.SR_Next;
  While (NextSequence<>StandByMechanism) And (NextSequence^.SR_Element <> Comp) Do
    Begin
    NextSequence:=NextSequence^.SR_Next;
    End;
  If NextSequence = StandByMechanism Then
    Begin
    NextSequence:=SequentialMechanism^.SR_Next;
    While (NextSequence^.SR_Element <> Comp) Do
      Begin
      NextSequence:=NextSequence^.SR_Next;
      End;
    End;
  If NextSequence = FutureEvent Then
    FutureEvent:=NextSequence^.SR_Next;  
  NextSequence^.SR_Pred^.SR_Next:=NextSequence^.SR_Next;
  NextSequence^.SR_Next^.SR_Pred:=NextSequence^.SR_Pred;
  Comp.FStatus:=TC_DATA;
  Comp.FEventTime:=Never;
  If NextSequence^.SR_LocalData <> Nil Then
    Begin
    FreeMem(NextSequence^.SR_LocalData,NextSequence^.SR_LocalData^.LD_Length+4);
    End;
  NextSequence^.SR_LocalData:=Nil;  
  Dispose(NextSequence);
//  NotifyServer;
  End;         {LeaveSequence}

procedure TomasProcessElement.RemoveProcess(Stop: Boolean);
  Begin
  If (SB_Successor <> Nil) And (SB_Successor^.SR_Element = Self) Then
    Begin
    SB_Successor:=SB_Successor^.SR_Next;
    End;
  LeaveSequence(Self);
  If Stop Then
    Begin
    FStatus:=TC_DATA;
    End
  Else
    Begin
    FStatus:=TC_PASSIVE;
    End;
  End;   {RemoveProcess}

Procedure TomasProcessElement.CancelResume;
  Var
    ELocData: LocalData;
  begin
  If Assigned(FLocData) Then
    begin
    ELocdata:=FLocData;
    FreeMem(ELocData,ELocData^.LD_Length+4);
    FLocData:=Nil;
    End;
  FStatus:=TC_DATA;
  End;

Procedure TomasProcessElement.InterruptProcess;
  Var
    NextSequence: Sequel;
  Begin
  If (SB_Successor <> Nil) And (SB_Successor^.SR_Element = Self) Then
    Begin
    SB_Successor:=SB_Successor^.SR_Next;
    End;
  NextSequence:=StandbyMechanism^.SR_Next;
  While (NextSequence<>StandByMechanism) And (NextSequence^.SR_Element <> Self) Do
    Begin
    NextSequence:=NextSequence^.SR_Next;
    End;
  If NextSequence = StandByMechanism Then
    Begin
    NextSequence:=SequentialMechanism^.SR_Next;
    While (NextSequence^.SR_Element <> Self) Do
      Begin
      NextSequence:=NextSequence^.SR_Next;
      End;
    End;
  If NextSequence = FutureEvent Then
    FutureEvent:=NextSequence^.SR_Next;
  NextSequence^.SR_Pred^.SR_Next:=NextSequence^.SR_Next;
  NextSequence^.SR_Next^.SR_Pred:=NextSequence^.SR_Pred;
  FStatus:=TC_INTERRUPTED;
  FEventTime:=Never;
  If NextSequence^.SR_LocalData <> Nil Then
    Begin
    FLocData:=NextSequence^.SR_LocalData;
    NextSequence^.SR_LocalData:=Nil;
    End;
  Dispose(NextSequence);
//  NotifyServer;
  End;   {InterruptProcess}

Procedure TomasProcessElement.CancelProcess;
  Var
    NextSequence: Sequel;
  Begin
  If (SB_Successor <> Nil) And (SB_Successor^.SR_Element = Self) Then
    Begin
    SB_Successor:=SB_Successor^.SR_Next;
    End;
  NextSequence:=StandbyMechanism^.SR_Next;
  While (NextSequence<>StandByMechanism) And (NextSequence^.SR_Element <> Self) Do
    Begin
    NextSequence:=NextSequence^.SR_Next;
    End;
  If NextSequence = StandByMechanism Then
    Begin
    NextSequence:=SequentialMechanism^.SR_Next;
    While (NextSequence^.SR_Element <> Self) Do
      Begin
      NextSequence:=NextSequence^.SR_Next;
      End;
    End;
  If NextSequence = FutureEvent Then
    FutureEvent:=NextSequence^.SR_Next;
  NextSequence^.SR_Pred^.SR_Next:=NextSequence^.SR_Next;
  NextSequence^.SR_Next^.SR_Pred:=NextSequence^.SR_Pred;
  FStatus:=TC_PASSIVE;
  FEventTime:=Never;
  If NextSequence^.SR_LocalData <> Nil Then
    Begin
    FLocData:=NextSequence^.SR_LocalData;
    NextSequence^.SR_LocalData:=Nil;
    End;
  Dispose(NextSequence);
//  NotifyServer;
  End;

Function CompInSequence(Comp: TomasProcessElement): Boolean;
  Var
    NextSequence: Sequel;
  Begin
  NextSequence:=StandbyMechanism^.SR_Next;
  While (NextSequence<>StandByMechanism) And (NextSequence^.SR_Element <> Comp) Do
    Begin
    NextSequence:=NextSequence^.SR_Next;
    End;
  If NextSequence = StandByMechanism Then
    Begin
    NextSequence:=SequentialMechanism^.SR_Next;
    While (NextSequence<>SequentialMechanism) And (NextSequence^.SR_Element <> Comp) Do
      Begin
      NextSequence:=NextSequence^.SR_Next;
      End;
    End;
  If NextSequence = SequentialMechanism Then
    Begin
    Result:=FALSE;
    End
  Else
    Begin
    Result:=TRUE;
    End;
  End;

Procedure HoldProcess(T: Double);
   Label
    RetHold;
{ebx contains address of Element,
 registers are on stack at entry;
 then a 4-byte returnaddress: so returnaddress is at dword[ebp + 4];
 then there is 8-byte double parameter T.
 stack will be saved in localdata and broken down until starting
 point of process }
  Begin
    asm
    push edi        {EDI must be saved!!!}
    push esi        {ESI also}
    push ebx
    push ebp
    mov ecx,InitialStack
    mov edx,esp    {current stackposition}
    sub ecx,esp    {total nr of bytes to save}
    mov NrOfBytes,ecx
    end;
  GetMem(LocData,NrOfBytes+4);
  LocData^.LD_Length:=NrOfBytes;
  StackPosition:=Addr(LocData^.LD_Data[1]);
    asm
    mov  eax,esi
    mov  edx,edi
    mov  edi,StackPosition
    mov  esi,esp
    mov  ecx,NrOfBytes
    rep  movsb                          {save bytes on stack}
    mov  ecx,OFFSET RetHold
    mov  JumpAddress,ecx
    mov  esi,eax
    mov  edi,edx
    end;
  TomasStatus.TS_CurrentElement.FReturnAddress:=JumpAddress;
  EnterSequence(TomasStatus.TS_Now + T,TomasStatus.TS_CurrentElement);
    asm
    mov esp,InitialStack     {breakdown stack}
    end;
  SelectNextCurrent;
    asm
  RetHold:
    end;
  End;

Procedure StandByProcess;
  Label
    RetStandBy;
  Begin
    asm
    push edi
    push esi
    push ebx
    push ebp
    mov ecx,InitialStack
    mov edx,esp    {current stackposition}
    sub ecx,esp    {total nr of bytes to save}
    mov NrOfBytes,ecx
    end;
  GetMem(LocData,NrOfBytes+4);
  LocData^.LD_Length:=NrOfBytes;
  StackPosition:=Addr(LocData^.LD_Data[1]);
    asm
    mov  eax,esi
    mov  edx,edi
    mov edi,StackPosition
    mov esi,esp
    mov ecx,NrOfBytes
    rep movsb                          {save bytes on stack}
    mov  ecx,OFFSET RetStandBy
    mov  JumpAddress,ecx
    mov  esi,eax
    mov  edi,edx
    end;
//  If FReturnAddress already is JumpAddress then returned from standby
//  so select next from standbychain
//  else new standby and then to end of standbychain
  If TomasStatus.TS_CurrentElement.FReturnAddress=JumpAddress Then
    Begin
    EnterStandBySequence(TomasStatus.TS_CurrentElement,FALSE);
    End
  Else
    Begin
    TomasStatus.TS_CurrentElement.FReturnAddress:=JumpAddress;
    EnterStandBySequence(TomasStatus.TS_CurrentElement,TRUE);
    End;
    asm
    mov esp,InitialStack     {breakdown stack}
    end;
  SelectNextCurrent;
    asm
  RetStandBy:
    end;
  End;

Procedure SuspendProcess;
  Label
    RetSuspend;
  Begin
    asm
    push edi
    push esi
    push ebx
    push ebp
    mov ecx,InitialStack
    mov edx,esp    {current stackposition}
    sub ecx,esp    {total nr of bytes to save}
    mov NrOfBytes,ecx
    end;
  GetMem(LocData,NrOfBytes+4);
  LocData^.LD_Length:=NrOfBytes;
  StackPosition:=Addr(LocData^.LD_Data[1]);
  TomasStatus.TS_CurrentElement.FLocData:=LocData;
    asm
    mov  eax,esi
    mov  edx,edi
    mov edi,StackPosition
    mov esi,esp
    mov ecx,NrOfBytes
    rep movsb                          {save bytes on stack}
    mov  ecx,OFFSET RetSuspend
    mov  JumpAddress,ecx
    mov  esi,eax
    mov  edi,edx
    end;
  TomasStatus.TS_CurrentElement.FReturnAddress:=JumpAddress;
  TomasStatus.TS_CurrentElement.FEventTime:=Never;
  LocData:=Nil;
    asm
    mov esp,InitialStack     {breakdown stack}
    end;
  SelectNextCurrent;
    asm
  RetSuspend:
    end;
  End;

Procedure FinishProcess;
  Begin
  TomasStatus.TS_CurrentElement.FStatus:=TC_DATA;
  SelectNextCurrent;
  End;

Procedure FinishProcessAndDestroy;
  Begin
  TomasStatus.TS_CurrentElement.FStatus:=TC_DATA;
  DestroyRoom.Add(TomasStatus.TS_CurrentElement);
  SelectNextCurrent;
  End;

Procedure StartProcessSimulation;
  Label
    LSimulationReturn;
  Begin
    asm
//    pop edx         {returnaddress}
//    mov SimulationReturn,edx
    mov InitialStack,esp
    mov StartEBP,ebp
    mov StartESI,esi
    mov StartEDI,edi
    mov StartEAX,eax
    mov StartEBX,ebx
    mov StartECX,ecx
    mov edx,offset LSimulationReturn
    mov SimulationReturn,edx
    end;
  TomasStatus.TS_Now:=0;
  If NOT(InStepMode) Then
    Begin
    TomasStatus.TS_SimulationStatus:=S_RUNNING;
    End;
  TomasStatus.TS_NextStep:=TRUE;
//  CheckServer;
  SelectNextCurrent;
    asm
  LSimulationReturn:
    end;
  End;

Procedure ResumeProcessSimulation;
  Begin
  If InStepMode Then
    Begin
    TomasStatus.TS_SimulationStatus:=S_STEPRUNNING;
    TomasStatus.TS_NextStep:=TRUE;
    End
  Else
    Begin
    TomasStatus.TS_SimulationStatus:=S_RUNNING;
    End;
  End;

Procedure InterruptProcessSimulation;
  Begin
  TomasStatus.TS_SimulationStatus:=S_INTERRUPTED;
  End;

Function CurrentProcessElement: TomasProcessElement;
  Begin
  Result:=TomasStatus.TS_CurrentElement;
  End;


(*
Procedure CheckServer;
  Begin
  If (TomasStatus.TS_ServerAddress<>'') And (ModelSocket.State <> wsConnected) Then
    Begin
    ModelSocket.Addr   := TomasStatus.TS_ServerAddress;        { Server host name }
    ModelSocket.Proto  := 'tcp';                { Protocol we want to use      }
    ModelSocket.Port   := '52623';             { The port we want to connect  }
    TomasStatus.TS_Tracing:='Connection called';
    StatusChange;
    ModelSocket.Connect;                      { Let's connect !               }
    GoOnModel:=FALSE;
    While (ModelSocket.State <> wsConnected) Or NOT(GoOnModel) Do
      Begin
      Application.ProcessMessages;
      End;
    End;
  End;

Function ConnectedToServer: Boolean;
  Begin
  Result:=(ModelSocket.State = wsConnected);
  End;

Constructor TModelSocket.Create(AOwner: TComponent);
  Begin
  inherited Create(AOwner);
  OnSessionConnected:=RegisterConnection;
  OnDataAvailable:=CheckServerAnswer;
  End;

procedure TModelSocket.RegisterConnection(Sender: TObject; Error: Word);
  begin
  If Error <> 0 Then
    Begin
    TomasStatus.TS_Tracing:='Connection failed';
    End
  else
    Begin
    TomasStatus.TS_Tracing:='Connected to Server';
    End;
  StatusChange;
  end;

procedure TModelSocket.CheckServerAnswer(Sender: TObject; Error: Word);
Var
  S: String;
  STot: String;
  Sm: String;
  Len: Integer;
  P: Integer;
  MsgNr: Integer;
  HL: Integer;
  TSequel: Sequel;
  NewReply: TReply;
  NewQuestion: TQuestion;

  Function GetMsgNr(Msg: String): Integer;
    Var
      i: Integer;
    Begin
    i:=10;
    While (I>0) And (Pos(ServerMessages[i],Msg)=0) Do
      Begin
      Dec(i);
      End;
    HL:=Length(ServerMessages[i]);
    Result:=i;
    End;

begin
Len := Receive(@Buffer[0], sizeof(Buffer) - 1);
While (Len > 0) Do
  Begin
  While (Len > 0) AND (Buffer[Len-1] in [#13, #10]) Do
    Begin
    Dec (Len);
    End;
  Buffer[Len] := #0;
  STot := Trim(StrPas(Buffer));
  While STot <> '' Do
    Begin
    P:=Pos(#13#10,STot);
    If P = 0 Then
      Begin
      S:=STot;
      STot:='';
      End
    Else
      Begin
      S:=Copy(STot,1,P-1);
      STot:=Copy(STot,P + 2, Length(STot) - P - 1);
      End;
    MsgNr:=GetMsgNr(S);
    Case MsgNr Of
      0:Begin    {unknown answer; send it to receiveproc}
        TomasStatus.TS_Message:=S;
        LogMessage(S,' received: ');
        StatusChange;
        End;
      1:Begin    {! = reply}
        //create a reply from the message
        LogMessage(S,' received: ');
        P:=Pos('!!',S);
        NewReply:=TReply.Create;
        NewReply.FID:=StrToInt(Copy(S,HL+1,P-HL-1));
        S:=Copy(S,P+2,Length(S) - P - 1);
        P:=Pos('!!',S);
        NewReply.FSource:=Copy(S,1,P-1);
        NewReply.FAnswer:=Copy(S,P+2,Length(S) - P - 1);
        ReplyList.Add(NewReply);
        NoReplyReceived:=FALSE;
        End;
      2:Begin    {? = question}
        LogMessage(S,' received: ');
        NewQuestion:=TQuestion.Create;
        P:=Pos('!!',S);
        NewQuestion.FID:=StrToInt(Copy(S,HL+1,P-HL-1));
        S:=Copy(S,P+2,Length(S) - P - 1);
        P:=Pos('!!',S);
        NewQuestion.FTarget:=Copy(S,1,P-1);
        NewQuestion.FQuestion:=Copy(S,P+2,Length(S) - P - 1);
        TomasStatus.TS_Question:=NewQuestion;
        StatusChange;
        End;
      3:Begin    {hello}
        P:=Pos('!!',S);
        ModelNumber:=Copy(S,HL+1,P - HL - 1);
        TomasStatus.TS_Now:=StrToFloat(Trim(Copy(S,P + 2,Length(S)- P - 1)));
        LogMessage(S,' received: ');
        TSequel:=SequentialMechanism^.SR_Next;
        While TSequel <> SequentialMechanism Do
          Begin
          If TSequel^.SR_Time <> NEVER Then
            Begin
            TSequel^.SR_Time:=TSequel^.SR_Time + TomasStatus.TS_Now;
            End;
          TSequel:=TSequel^.SR_Next;
          End;
        TomasSendStr('$$name:' + ModelName + #13#10);
        GoOnModel:=TRUE;
        End;
      4:Begin    {continue}
        TomasStatus.TS_Now:=StrToFloat(Trim(Copy(S,HL+1,Length(S)-HL)));
        LogMessage(S,' received: ');
        GoOnModel:=TRUE;
        End;
      5:Begin    {stop}
        SimulationStopTime:=StrToFloat(Trim(Copy(S,HL+1,Length(S)-HL)));
        LogMessage(S,' received: ');
        GoOnModel:=TRUE;
        SelectNextCurrent;
        End;
      6:Begin    {Simtime}
        TomasStatus.TS_Now:=StrToFloat(Trim(Copy(S,HL+1,Length(S)-HL)));
        LogMessage(S,' received: ');
        TomasStatus.TS_Message:=S;
        StatusChange;
        End;
      7:Begin    {time}
        TomasStatus.TS_Message:=S;
        LogMessage(S,' received: ');
        StatusChange;
        End;
      8:Begin    {clients}
        TomasStatus.TS_Message:=S;
        LogMessage(S,' received: ');
        StatusChange;
        End;
      9:Begin    {from}
        P:=Pos('!!',S);
        Sm:=Copy(S,HL+1,P - HL -1);
        S:=Copy(S,1,HL) + Copy(S,P+2,Length(S) - P - 1);
        If Synchronization then
          TomasStatus.TS_Now:=StrToFloat(Trim(Sm));
        TomasStatus.TS_Message:=S;
        LogMessage(S,' received: ');
        StatusChange;
        End;
      10:Begin    {unknown}
        TomasStatus.TS_Message:=S;
        LogMessage(S,' received: ');
        StatusChange;
        End;
      End;
    End;
  Len := Receive (@Buffer[0], sizeof (Buffer) - 1);
  End;
end;

Procedure DisconnectFromServer;
  Begin
  If (TomasStatus.TS_ServerAddress<>'') And (ModelSocket.State = wsConnected) Then
    Begin
    ModelSocket.Close;
    GoOnModel:=TRUE;
    End;
  End;
*)





Procedure DoStartLogging;
  Begin
  if LoggingStarted Then
    Begin
    If FileExists(ModelName + '.log') Then
      Begin
      Append(LogFile);
      End
    Else
      Begin
      AssignFile(LogFile,ModelName + '.log');
      ReWrite(LogFile);
      LogCount:=0;
      End;
    WriteLn(LogFile,'***' + FormatFloat('0.00',GetNow) + ' resumes logging');
    CloseFile(LogFile);
    End
  Else
    begin
    AssignFile(LogFile,ModelName + '.log');
    ReWrite(LogFile);
    WriteLn(LogFile,'***' + FormatFloat('0.00',GetNow) + ' start logging');
    CloseFile(LogFile);
    LogCount:=0;
    LoggingStarted:=TRUE;
    End;
  Logging:=TRUE;
  Inc(LogCount);
  End;

Procedure DoStopLogging;
  Begin
  If Logging Then
    Begin
    Append(LogFile);
    WriteLn(LogFile,'***' + FormatFloat('0.00',GetNow) + ' stop logging');
    CloseFile(LogFile);
    Inc(LogCount);
    End;
  End;

Function GetLogging: Boolean;
  Begin
  Result:=Logging;
  End;

Procedure SetProcessingInterval(PInt: Integer);
  Begin
  ProcessingInterval:=PInt;
  EventCounter:=PInt;
  End;  

(*
Procedure DoSendQuestionTo(Model: String; Var Question: String);
  Var
    NewQuestion: TQuestion;
    TheReply: TReply;
    AnswerToThisQuestion: Boolean;
    I: Integer;
  Begin
  Inc(QuestionID);
  If Model = '' Then
    begin
    Model:='server';
    End;
  NewQuestion:=TQuestion.Create;
  NewQuestion.FID:=QuestionID;
  NewQuestion.FTarget:=Model;
  NewQuestion.FQuestion:=Question;
  QuestionList.Add(NewQuestion);
  TomasSendStr('$$?' + IntToStr(NewQuestion.FID) + '!!' + NewQuestion.FTarget
   + '!!' + NewQuestion.FQuestion + #13#10);
  AnswerToThisQuestion:=FALSE;
  TheReply:=Nil;
  While NOT(AnswerToThisQuestion) Do
    Begin
    While NoReplyReceived Do
      Begin
      Application.ProcessMessages;
      End;
    I:=0;
    While (I<=ReplyList.Count - 1) And Not(AnswerToThisQuestion) Do
      Begin
      TheReply:=ReplyList.Items[i];
      If TheReply.FID=NewQuestion.FID Then
        Begin
        AnswerToThisQuestion:=TRUE;
        End
      Else
        Begin
        Inc(i);
        End;
      End;
    End;
  If AnswerToThisQuestion Then
    Begin
    QuestionList.Remove(NewQuestion);
    NewQuestion.Free;
    Question:=TheReply.FAnswer;
    ReplyList.Remove(TheReply);
    TheReply.Free;
    NoReplyReceived:=TRUE;
    End;
  End;

Procedure DoSendReplyTo(Model,Reply: String; ID: Integer);
  Begin
  TomasSendStr('$$!' + IntToStr(ID) + '!!' + Model + '!!' + Reply + #13#10);
  End;

*)



Procedure ResetTomasProcess;
  Var
    TheSequel: Sequel;
  Begin
  TheSequel:=SequentialMechanism^.SR_Next;
  While TheSequel <> SequentialMechanism Do
    Begin
    TheSequel^.SR_Pred^.SR_Next:=TheSequel^.SR_Next;
    TheSequel^.SR_Next^.SR_pred:=TheSequel^.SR_Pred;
    Dispose(TheSequel);
    TheSequel:=SequentialMechanism^.SR_Next;
    End;
  TheSequel:=StandByMechanism^.SR_Next;
  While TheSequel <> StandByMechanism Do
    Begin
    TheSequel^.SR_Pred^.SR_Next:=TheSequel^.SR_Next;
    TheSequel^.SR_Next^.SR_pred:=TheSequel^.SR_Pred;
    Dispose(TheSequel);
    TheSequel:=StandByMechanism^.SR_Next;
    End;
  FutureEvent:=SequentialMechanism;
  InStandBy:=FALSE;
  SB_Successor:=Nil;
  DestroyRoom.Clear;
  SimulationStopTime:=Never;
  SimulationBreakTime:=-1;
  SimulationKeyTime:=-1;
  //ModelSocket.Close;
  TomasStatus.TS_SimulationStatus:=S_INIT;
  TomasStatus.TS_CurrentElement:=Nil;
  TomasStatus.TS_Now:=0;
  TomasStatus.TS_Tracing:='';
  TomasStatus.TS_Message:='';
  End;  {ResetTomasProcess}


Procedure InitializeTomasProcess;
  Begin
  New(SequentialMechanism);
  SequentialMechanism^.SR_Next:=SequentialMechanism;
  SequentialMechanism^.SR_Pred:=SequentialMechanism;
  SequentialMechanism^.SR_Element:=Nil;
  SequentialMechanism^.SR_Time:=NEVER;
  SequentialMechanism^.SR_StandBy:=FALSE;
  SequentialMechanism^.SR_Localdata:=Nil;
  New(StandByMechanism);
  StandByMechanism^.SR_Next:=StandByMechanism;
  StandByMechanism^.SR_Pred:=StandByMechanism;
  StandByMechanism^.SR_Element:=Nil;
  StandByMechanism^.SR_Time:=NEVER;
  StandByMechanism^.SR_StandBy:=TRUE;
  StandByMechanism^.SR_Localdata:=Nil;
  FutureEvent:=SequentialMechanism;
  InStandBy:=FALSE;
  SB_Successor:=Nil;
  DestroyRoom:=TList.Create;
  QuestionList:=TList.Create;
  ReplyList:=TList.Create;
  QuestionID:=0;
  NoReplyReceived:=TRUE;
  ReplyRecognized:=FALSE;
  TomasStatus:=TTomasStatus.Create;
  SimulationStopTime:=Never;
  SimulationBreakTime:=-1;
  SimulationKeyTime:=-1;
  InStepMode:=FALSE;
//  ModelSocket:=TModelSocket.Create(Nil);
  ModelName:='TomasModel';
//  WaitMsg:='';
  GoOnModel:=FALSE;
  Logging:=FALSE;
  LoggingStarted:=FALSE;
  ProcessingInterval:=1;
  EventCounter:=1;
  Synchronization:=TRUE;
  End;

Procedure FinishTomasProcess;
  Begin
  ResetTomasProcess;
  Dispose(SequentialMechanism);
  DestroyRoom.Free;
  TomasStatus.Free;
//  ModelSocket.Free;
  End;

Initialization
  InitializeTomasProcess;

Finalization
  FinishTomasProcess;

end.
