unit TomasRead;

Interface

Uses
  Classes,
  Windows,
  SysUtils,
  Dialogs,
  ExtCtrls,
  StdCtrls;

Type
  TomasFile = class
    Private
    FName: String;
    FFile: TextFile;
    FOpen: Boolean;
    FText: TStringList;
    FDecimalSeparator: Char;
    FFileMarker: String;
    FFileMarkerRead: Boolean;
    FSubString: String;
    Function Open: Boolean;
    Procedure SetNextSubString;
    Public
    Constructor Create(FNm: String);
    Destructor Destroy; Override;
    Function GetBoolean: Boolean;   {TRUE if y,Y,t,T,1}
    Function GetSmallInt: SmallInt; {-32768,..,32767}
    Function GetLongInt: LongInt;   {-2147483648,..,2147483647}
    Function GetByte: Byte;         {0,..,255}
    Function GetWord: Word;         {0,..,65535}
    Function GetInteger: Integer;   {-2147483648,..,2147483647}
    Function GetCardinal: Cardinal; {0,..,2147483647}
    Function GetReal: Real;         {6 bytes}
    Function GetDouble: Double;     {8 bytes}
    Function GetSingle: Single;     {4 bytes}
    Function GetString: String;
    Function FileMarker: Boolean;
    End;


implementation

Constructor TomasFile.Create(FNm: String);
  Begin
  Inherited Create;
  FName:=FNm;
  FOpen:=FALSE;
  FText:=TStringList.Create;
  FDecimalSeparator:=DecimalSeparator;
  FFileMarker:='//';
  FFileMarkerRead:=FALSE;
  End; {TomasFile.Create}

Destructor TomasFile.Destroy;
  Begin
  If FOpen Then
    Begin
    CloseFile(FFile);
    End;
  FText.Free;
  Inherited Destroy;
  End; {TomasFile.Destroy}

Function TomasFile.Open: Boolean;
  Var
    i: Integer;
    j: Integer;
    k: Integer;
    S: String;
    CommentOpen: Boolean;
  Begin
  If NOT(FOpen) Then
    Begin
    If FileExists(FName) Then
      Begin
      AssignFile(FFile,FName);
      {$I-}
      Reset(FFile);
      {$I+}
      i:=IOResult;
      If i <> 0 Then
        Begin
        MessageDlg('Error: Could not open File ' + FName
         + '. I/O-error = ' + IntToStr(i),mtError,[mbOK],0);
        RunError(i);
        End
      Else
        Begin
        FOpen:=TRUE;
        End;
      End
    Else
      Begin
      MessageDlg('Error: File ' + FName + ' does not exist.',
        mtError,[mbOK],0);
      RunError(255);
      End;
    End;
  If (FText.Count = 0) Or FileMarker Then
    Begin
    If EOF(FFile) Then
      Begin
      Result:=FALSE;
      Exit;
      End
    Else
      Begin
      FFileMarkerRead:=FALSE;
      FText.Clear;
      While NOT(EOF(FFile)) And NOT(FFileMarkerRead) Do
        Begin
        ReadLn(FFile,S);
        If Trim(S) <> '' Then   {skip empty lines}
          Begin
          FText.Append(S);
          If Trim(S) = '//' Then
            Begin
            FFileMarkerRead:=TRUE;
            End;
          End;
        End;
      End;
    i:=-1;
    CommentOpen:=FALSE;
    While i < FText.Count - 1 Do
      Begin
      Inc(i);
      If CommentOpen Then
        Begin
        k:=Pos('}',FText.Strings[i]);
        If k = 0 Then
          Begin
          FText.Strings[i]:='';
          End
        Else
          Begin
          CommentOpen:=FALSE;
          FText.Strings[i]:=Copy(FText.Strings[i],k+1,Length(FText.Strings[i]) - k);
          End;
        End;
      j:=Pos('{',FText.Strings[i]);
      While (j <> 0) And NOT(CommentOpen) Do
        Begin
        CommentOpen:=TRUE;
        k:=Pos('}',FText.Strings[i]);
        If k = 0 Then
          Begin
          FText.Strings[i]:=Copy(FText.Strings[i],1,j-1);
          End
        Else
          Begin
          CommentOpen:=FALSE;
          FText.Strings[i]:=Copy(FText.Strings[i],1,j - 1)
           + Copy(FText.Strings[i],K+1,Length(FText.Strings[i])-k);
          End;
        j:=Pos('{',FText.Strings[i]);
        End;
      FText.Strings[i]:=Trim(FText.Strings[i]);
      If FText.Strings[i] = '' Then
        Begin
        FText.Delete(i);
        Dec(i);
        End;
      End;
    End;
  Result:=TRUE;
  End; {TomasFile.Open}

Procedure TomasFile.SetNextSubString;
  Var
    InString: Boolean;
    i: Integer;
  Begin
  If Open Then
    Begin
    FSubString:='';
    i:=1;
    InString:=FALSE;
    While (i <= Length(FText.Strings[0])) And
     ((Copy(FText.Strings[0],i,1) <> ' ') Or (InString)) Do
      Begin
      If Copy(FText.Strings[0],i,1) = '''' Then
        Begin
        If InString Then
          Begin
          If (Length(FText.Strings[0]) > i)
           And (Copy(FText.Strings[0],i+1,1) = '''') Then
            Begin
            FSubString:=FSubString + '''';
            Inc(i);
            End
          Else
            Begin
            InString:=FALSE;
            End;
          End
        Else
          Begin
          InString:=TRUE;
          End;
        End
      Else
        Begin
        FSubString:=FSubString + Copy(FText.Strings[0],i,1);
        End;
      Inc(i);
      End;
    If I > Length(FText.Strings[0]) Then
      Begin
      FText.Delete(0);
      End
    Else
      Begin
      FText.Strings[0]:=Trim(Copy(FText.Strings[0],i+1,Length(FText.Strings[0]) - i));
      End;
    End
  Else
    Begin
    FSubString:='$$$TomasEmpty';
    End;
  End;   {TomasFile.SetNextSubString}

Function TomasFile.GetBoolean: Boolean;
  Var
    C: Char;
  Begin
  SetNextSubString;
  C:=FSubString[1];
  Case C Of
    'y','Y','t','T','1':
      Begin
      Result:=TRUE;
      End;
    Else
      Begin
      Result:=FALSE;
      End;
    End;
  End;  {TomasFile.GetBoolean}

Function TomasFile.GetSmallInt: SmallInt;
  Var
    II: LongInt;
  Begin
  II:=GetLongInt;
  If II = 2147483647 Then
    Begin
    Result:=32767;
    End
  Else
    Begin
    If (II>32767) Or (II<-32768) Then
      Begin
      Result:=32767;
      End
    Else
      Begin
      Result:=II;
      End;
    End;
  End;  {TomasFile.GetSmallInt}

Function TomasFile.GetLongInt: LongInt;
  Var
    I: LongInt;
  Begin
  SetNextSubString;
  If FSubString = '$$$TomasEmpty' Then
    begin
    Result:=2147483647;
    End
  Else
    Begin
    Try
      I:=StrToInt(FSubString);
    Except
      On E: EConvertError Do I:=2147483647;
      End;
    Result:=I;
    End;
  End;  {TomasFile.GetLongInt}

Function TomasFile.GetByte: Byte;
  Var
    II: LongInt;
  Begin
  II:=GetLongInt;
  If II = 2147483647 Then
    Begin
    Result:=255;
    End
  Else
    Begin
    If (II>255) Or (II<0) Then
      Begin
      Result:=255;
      End
    Else
      Begin
      Result:=II;
      End;
    End;
  End;  {TomasFile.GetByte}

Function TomasFile.GetWord: Word;
  Var
    II: LongInt;
  Begin
  II:=GetLongInt;
  If II = 2147483647 Then
    Begin
    Result:=65535;
    End
  Else
    Begin
    If (II>65535) Or (II<0) Then
      Begin
      Result:=65535;
      End
    Else
      Begin
      Result:=II;
      End;
    End;
  End;  {TomasFile.GetWord}

Function TomasFile.GetInteger: Integer;
  Begin
  Result:=GetLongInt;
  End;  {TomasFile.GetInteger}

Function TomasFile.GetCardinal: Cardinal;
  Var
    II: LongInt;
  Begin
  II:=GetLongInt;
  If (II<0) Then
    Begin
    Result:=2147483647;
    End
  Else
    Begin
    Result:=II;
    End;
  End;  {TomasFile.GetCardinal}

Function TomasFile.GetReal: Real;
  var
    D: Double;
  Begin
  D:=GetDouble;
  If (D > 1.7E+38) Or (D < -1.7E+38) Then
    Begin
    Result:=1.7E+38;
    End
  Else
    begin
    Result:=D;
    End;
  End;  {TomasFile.GetReal}

Function TomasFile.GetDouble: Double;
  Var
    D: Double;
  Begin
  SetNextSubString;
  If FSubString = '$$$TomasEmpty' Then
    begin
    Result:=1.7E+308;
    End
  Else
    Begin
    Try
      D:=StrToFloat(FSubString);
    Except
      D:=1.7E+308;
      End;
    Result:=D;
    End;
  End;  {TomasFile.GetDouble}

Function TomasFile.GetSingle: Single;
  Var
    S: Single;
  Begin
  SetNextSubString;
  If FSubString = '$$$TomasEmpty' Then
    begin
    Result:=1.7E+308;
    End
  Else
    Begin
    Try
      S:=StrToFloat(FSubString);
    Except
      S:=3.3E+38;
      End;
    Result:=S;
    End;
  End;  {TomasFile.GetSingle}

Function TomasFile.GetString: String;
  Begin
  SetNextSubString;
  Result:=FSubString;
  End; {TomasFile.GetString}

Function TomasFile.FileMarker: Boolean;
  Begin
  Result:=FALSE;
  If NOT(FOpen) Then
    Begin
    Open;
    End;
  If FOpen Then
    Begin
    If (FText.Count > 0) And (Length(FText.Strings[0]) > 0)
     And (Copy(FText.Strings[0],1,2) = FFileMarker) Then
      Begin
      FText.Delete(0);
//      FText.Clear;
      Result:=TRUE;
      End
    Else
      Begin
      Result:=FALSE;
      End;
    End;
  End;  {TomasFile.FileMarker}

end.
